package genius.core.uncertainty;

import java.util.ArrayList;
import java.util.List;

import genius.core.boaframework.SortedOutcomeSpace;
import genius.core.utility.AbstractUtilitySpace;
import genius.core.utility.AdditiveUtilitySpaceWithPerturbation;
import genius.core.utility.UncertainAdditiveUtilitySpace;

/**
 * 
 * Keeps track of both the real preferences of the user as well as the
 * perception of the agent. The preference profile that is given to the agents
 * can be a "distorted utility space" in the cases that uncertainty is included
 * in the form of a perturbation to the real utilities (see
 * {@link AdditiveUtilitySpaceWithPerturbation}}) or a flattened Utility Space
 * (see {@link FlattenedUtilitySpace}), OR in the form of a Pairwise Comparison
 * User Model (see {@link UserModel}). Used only in the core.
 *
 */

public class UncertainPreferenceContainer {

	private UNCERTAINTYTYPE type;
	private UncertainAdditiveUtilitySpace realUtilitySpace;
	private UserModel pairwiseCompUserModel;

	/**
	 * 
	 * @param realUtilitySpace
	 * @param type
	 */
	public UncertainPreferenceContainer(UncertainAdditiveUtilitySpace utilspace,
			UNCERTAINTYTYPE type) {
		this.realUtilitySpace = utilspace;
		this.type = type;
		if (utilspace.isExperimental())
			createExperimentalOutcomeComparisonUserModel(
					utilspace.getComparisons(), utilspace.getErrors());
		else
			createOutcomeComparisonUserModel(utilspace.getComparisons(),
					utilspace.getErrors());
	}

	public UncertainPreferenceContainer(
			UncertainPreferenceContainer container) {
		this.realUtilitySpace = container.realUtilitySpace;
		this.type = container.type;
		this.pairwiseCompUserModel = container.pairwiseCompUserModel;
	}

	public UncertainPreferenceContainer(ExperimentalUserModel userModel) {
		this.realUtilitySpace = userModel.getRealUtilitySpace();
		this.type = UNCERTAINTYTYPE.PAIRWISECOMP;
		this.pairwiseCompUserModel = userModel;
	}

	/**
	 * Generates comparisons between different outcomes. At the moment each
	 * outcome is part of only one comparison.
	 * 
	 * 
	 * @param amountOfComparisons,
	 *            amount of comparisons
	 * @error the number of errors. UNUSED
	 * 
	 */
	public void createOutcomeComparisonUserModel(int amountOfComparisons,
			double error) {
		SortedOutcomeSpace outcomeSpace = (new SortedOutcomeSpace(
				this.getRealUtilitySpace()));
		BidRanking ranking = (new ComparisonGenerator(outcomeSpace))
				.generateRankingByAmount(amountOfComparisons + 1);
		this.pairwiseCompUserModel = new UserModel(ranking);
	}

	// TODO REMOVE LATER, ReFactor for copy-paste
	public void createExperimentalOutcomeComparisonUserModel(
			int amountOfComparisons, double error) {
		SortedOutcomeSpace outcomeSpace = (new SortedOutcomeSpace(
				this.getRealUtilitySpace()));
		List<OutcomeComparison> bidComparisons = new ArrayList<OutcomeComparison>();
		BidRanking ranking = (new ComparisonGenerator(outcomeSpace))
				.generateRankingByAmount(amountOfComparisons + 1);
		this.pairwiseCompUserModel = new ExperimentalUserModel(ranking,
				realUtilitySpace);
	}

	public AbstractUtilitySpace getRealUtilitySpace() {
		return realUtilitySpace;
	}

	public UserModel getPairwiseCompUserModel() {
		return pairwiseCompUserModel;
	}

	public void setPairwiseCompUserModel(UserModel pairwiseCompUserModel) {
		this.pairwiseCompUserModel = pairwiseCompUserModel;
	}

	public UNCERTAINTYTYPE getType() {
		return type;
	}
}
